require 'class_extensions'


class Guidepost < ActiveRecord::Base
  include CommonGuidepost
  extend CommonGuidepost::ClassMethods
  
  has_many :signs, inverse_of: :guidepost, dependent: :destroy
  has_many :trails, through: :signs, inverse_of: :guideposts
#  validates_presence_of :kctref, :osmid # dříve bylo ale uz se nehodi 
  has_one :osm_guidepost, :foreign_key => :osmid, :primary_key=> :osmid, inverse_of: :guidepost
  has_many :photos, inverse_of: :guidepost
  belongs_to :sez_guidepost, inverse_of: :guidepost
  
  scope :without_osm_guidepost, -> { where(osm_guidepost: nil) }
  scope :without_position, -> { where(lat: nil, lon:nil)}
  scope :with_position, -> { where('lat IS NOT NULL AND lon IS NOT NULL') }
  scope :search, lambda { |term| where("CONCAT_WS('', kctref, LOWER(nazev)) LIKE ?", "%#{term.mb_chars.downcase
}%")  }
  
  # creation
  def initialize(options={})
    super
    self.nazev_normalized = self.nazev.normalize unless self.nazev.nil?
    self.copy_coordinates_from_osm_if_needed
  end
  def self.new_from_kct_guidepost(kct_guidepost)
    Guidepost.new(:nazev=> kct_guidepost.nazev,:kctref=> kct_guidepost.kctref, :vyska=> kct_guidepost.vyska)
  end
 
  def self.all_with_coordinates
    all.reject { |gp| gp.lat.nil? or gp.lon.nil? }
  end
  
  # převody na string vyjádření
  def nazev_a_kctref
    if nazev.blank? and osm_guidepost.andand.nazev.blank?
      kctref.blank? ? "dbf:#{id}" : kctref
    else
      nazev || osm_guidepost.andand.nazev + (kctref.nil? ? '' : " (#{kctref})") 
    end
  end
  def to_s(option=nil)
    case option
      when :long then nazev_a_kctref
      else "dbf:#{id}"
    end
  end
  def position_string
    "#{lat}, #{lon}"
  end
  
  # další
  def is_on?(trail)
    Sign.exists?(:guidepost => self, :trail => trail)
  end
   
  def coordinates; [lat,lon] end
  
  def self.merge_reasons_and_score(kct_guidepost,osm_guidepost)
    reasons = []; score = 0
  
    # by associations or references
    if !kct_guidepost.guidepost.nil? and (kct_guidepost.guidepost.id == osm_guidepost.guidepost.andand.id)
      score += 250
      reasons << 'Guideposts are alrady associated.'
    else
      if kct_guidepost.kctref == osm_guidepost.kctref
        score += 100
        reasons << 'Equal kctrefs.'
      end
    end
    
    # by names
    kgn = kct_guidepost.nazev_normalized.gsub(/[^[[:alnum:]]]/,'') 
    ogn = osm_guidepost.nazev_normalized.gsub(/[^[[:alnum:]]]/,'')
    
    if kgn==ogn 
      score += 75; reasons << 'Names are identical.'
    elsif kgn.start_with?(ogn) or ogn.start_with?(kgn)
      score += 65; reasons << 'One name is prefix of the second one.'
    else
      levenshtein_distance = Text::Levenshtein.distance(kct_guidepost.nazev_normalized,osm_guidepost.nazev_normalized || '')
      score += 
        case levenshtein_distance
          when 0 then 70
          when 1 then 55
          when 2 then 25
          else 0
        end
      reasons << "Levenshtein distance of names is small (#{levenshtein_distance})"  if levenshtein_distance<=2
    end
    
    [reasons,score]
  end
      
  ## nástroje KČT
  
  def exists_in_kct_by_nazev?;  KctGuidepost.find_by_nazev(self.nazev) rescue false end
  def exists_in_kct_by_kctref?; KctGuidepost.find_by_kctref(self.kctref) rescue false end
  def exists_in_kct?;           exists_in_kct_by_kctref? or exists_in_kct_by_nazev? end
  def kct_guidepost; kct_guideposts.first end
  def kct_guideposts; KctGuidepost.where(kctref: kctref.andand.split(';') || []).order(:kctref) end
  def title; !nazev.blank? ? "#{nazev} (#{kctref})" : "rozcestník #{kctref}" end
    
  def self.nearby(coordinates, max_meters)
    return [] if coordinates[0].nil? or coordinates[1].nil?
    Guidepost.where("lat BETWEEN #{coordinates[0]-max_meters/Map.min_dlat} AND #{coordinates[0]+max_meters/Map.min_dlat}")
      .where("lon BETWEEN #{coordinates[1]-max_meters/Map.min_dlon} AND #{coordinates[1]+max_meters/Map.min_dlon}")
      .collect { |guidepost| 
        [Map.distance(coordinates, guidepost.coordinates), guidepost]
      }.sort_by{|pair| pair[0] }.reject{ |pair| pair[0]>max_meters}
  end  
  
  def copy_coordinates_from_osm_if_needed
    if !osm_guidepost.nil?
      self.lat = osm_guidepost.lat if lat.blank?
      self.lon = osm_guidepost.lon if lon.blank?
    end
  end
  def osm_guideposts_nearby(max_meters)
    return [] if coordinates[0].nil? or coordinates[1].nil?
    OsmGuidepost.nearby([self.lat,self.lon], max_meters)
  end
  
  def self.symbol; '&#10500;'.html_safe end
  def self.find_by_kctref(kctref)
    if kctref.blank?
      return nil
    else
      return Guidepost.where("kctref REGEXP '(^|;)#{kctref}($|;)'", ).first
    end
  end
  def add_kctref(akctref)
    self.kctref = self.kctref + ';' + akctref
    self
  end
  def kctrefs
    (self.kctref || '').split(';')
  end
  def good_kctrefs
    kctrefs.select{ |kctref| KctGuidepost.looks_like_kctref? kctref }
  end
  ## zjistí zda jeden z kctref rozcestníku je akctref - respektuje přitom že rozcestník může mít víc kctref
  def has_kctref?(akctref)
    kctrefs.include? akctref
  end
  
  def zones_prefixes
    good_kctrefs.collect{ |kctref| kctref[0..1] }
  end
  def zones
    zones_prefixes.collect{ |prefix| Zone.find(prefix) }
  end
  
  # vrací hash s osm tagy, které reprezentují informace v guidepostu obsažené
  def suggested_osm_tags
    r = {}
    r['name'] = nazev unless nazev.blank?
    r['short_name'] = nazev_short unless nazev_short.blank?
    r['ref'] = kctref  unless kctref.blank?
    r['ele'] = vyska  unless vyska.blank?
    r['information'] = 'guidepost'
    r['tourism'] = 'information'
    r['hiking'] = 'yes'
    return r
  end
  
  def guidepost; self end # helps in templates
end

